-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.enableFFB = true --false

-- The direction of the real wheel's FFB
M.wheelFFBaxisdir = 1

-- The FFB Force coefficient
M.wheelFFBForceCoef = 400

-- The FFB steady force limits (in Newton)
M.wheelFFBForceLimit = 2

-- FFB smoothing
M.wheelFFBSmoothing = 100

-- High frequency FFB smoothing
M.wheelFFBSmoothingHF = 50

M.hydros = {}
M.lastFFBforce = 0
M.smoothFFBforce = 0

local FFBsmooth = newExponentialSmoothing(M.wheelFFBSmoothing)
local FFBHydros = {}
local FFBHydrosExist = false
local FFBID    = -1 -- >=0 are valid IDs
local wheelpos = 0
local FFBIntegral = 0
local curForceLimitSmoother = newExponentialSmoothing(100)
local curForceLimit = 0
local FFBForce = 0
local FFmax = 10
local FFstep = FFmax / 256
local smootherFFB = newExponentialSmoothing(2000)

local function nop()
end

local function toInputSpace(h, state)
    if state > h.center then
        return (state - h.center) * h.invMultOut
    else
        return (state - h.center) * h.invMultIn
    end
end

local function FFquantum(f)
    return math.floor(f/FFstep)
end

local function updateGFX(dt)
    -- update the source command value
    for i,h in pairs(M.hydros) do
        h.cmd = math.min(math.max(electrics.values[h.inputSource] or 0, h.inputInLimit), h.inputOutLimit) * h.inputFactor

        if h.cmd ~= h.inputCenter or h.analogue == true then
            h._inrate = h.inRate
            h._outrate = h.outRate
        else
            -- set autocenter rate
            h._inrate = h.autoCenterRate
            h._outrate = h.autoCenterRate
        end

        if h.cmd >= h.inputCenter then
            h.cmd = h.cOut + h.cmd * h.multOut
        else
            h.cmd = h.cIn + h.cmd * h.multIn
        end

        h.hydroDirState = toInputSpace(h, h.state)
    end

    if FFBHydrosExist then
        curForceLimit = curForceLimitSmoother:get(M.wheelFFBForceLimit)
        wheelpos = electrics.values["steering_input"] or 0
        if FFquantum(FFBForce) ~= FFquantum(M.lastFFBforce) then
            M.lastFFBforce = FFBForce
            if FFBID >= 0 and vehicleMode == 1 then
                obj:sendForceFeedback(FFBID, M.wheelFFBaxisdir * M.lastFFBforce)
            end
        end
    end
end

local function update(dts)
    -- state: the state of the hydro from -1 to 1
    -- cmd the input value
    -- note: state is scaled to the ratio as the last step
    local hcmd = -1
    for i,h in pairs(M.hydros) do
        -- slowly approach the desired value
        if h.cmd < h.state then
            h.state = math.max(h.state - dts * h._inrate, h.cmd)
        else
            h.state = math.min(h.state + dts * h._outrate, h.cmd)
        end

        --if inputSource ~= "steering" then
        --    log('D', "hydros.update", "state: "..h.state .. " - inlimit: " .. h.inLimit.. " - state: "..h.state .. " cmd:"..h.cmd .. " val = "..val)
        --end
        obj:setBeamLengthRefRatio(h.beam.cid, h.state)
    end

    if FFBHydrosExist then
        local hcount = 0
        local hpos = 0
        for i,h in pairs(FFBHydros) do
            local hbcid = h.beam.cid
            if not obj:beamIsBroken(hbcid) then
                hcount = hcount + 1
                hpos = hpos + toInputSpace(h, obj:getBeamLength(hbcid) * h.invFFBHydroRefL)
            end
        end

        if hcount == 0 then
            M.lastFFBforce = 0
        else
            local hposraw = hpos / hcount
            local wheeldispl = wheelpos - hposraw
            FFBForce = M.wheelFFBForceCoef * wheeldispl
            M.smoothFFBforce = smootherFFB:get(FFBForce)
            FFBForce = FFBsmooth:get(sign(FFBForce) * math.min(math.abs(FFBForce), curForceLimit))

            if sign(wheeldispl) ~= sign(M.lastFFBforce) then
                FFBWindow = M.wheelFFBSmoothingHF
            else
                FFBWindow = M.wheelFFBSmoothing
            end
            if math.abs(FFquantum(FFBForce) - FFquantum(M.lastFFBforce)) > 1 then
                M.lastFFBforce = FFBForce
                if FFBID >= 0 and vehicleMode == 1 then
                    obj:sendForceFeedback(FFBID, M.wheelFFBaxisdir * M.lastFFBforce)
                end
            end
            FFBsmooth:setWindow(FFBWindow)
        end
    end
	
	if electrics.values['brake_input'] > 0 then 
		electrics.values['pitch_input'] = math.min(1, electrics.values['pitch_input']+0.0005)
	end
	if electrics.values['throttle_input'] > 0 then 
		electrics.values['pitch_input'] = math.max(-1, electrics.values['pitch_input']-0.0005)
	end
	
	if electrics.values['throttle_input'] == 0 and electrics.values['brake_input'] == 0 then 
		if electrics.values['pitch_input'] > 0 then electrics.values['pitch_input'] = math.max(0, electrics.values['pitch_input']-0.0005) end
		if electrics.values['pitch_input'] < 0 then electrics.values['pitch_input'] = math.min(0, electrics.values['pitch_input']+0.0005) end
	end
	
	electrics.values['pitch_input'] = -electrics.values['throttle_input'] + electrics.values['brake_input']

    if not drivetrain.engineDisabled and v.data.engine and v.data.engine.propPresent then
    	if input.keys.T then
    		electrics.values['custom_throttle'] = math.min( 1, electrics.values['custom_throttle'] + 0.0003)
    	end
    	if input.keys.G then
    		electrics.values['custom_throttle'] = math.max( 0, electrics.values['custom_throttle'] - 0.0003)
    	end
    	obj:applyTorque(nd_tr1, nd_tr2, nd_tr3, -3000*electrics.values['custom_throttle']-20)
    	obj:applyTorque(nd_tr1, nd_tr3, nd_tr4, -3000*electrics.values['custom_throttle']-20)
    	obj:applyTorque(nd_tr1, nd_tr4, nd_tr2, -3000*electrics.values['custom_throttle']-20)
	end
    
	electrics.values['e_r'] = -electrics.values['steering_input'] - electrics.values['pitch_input']
	electrics.values['e_l'] = electrics.values['steering_input'] - electrics.values['pitch_input']
end

-- nop'ed functions
M.updateGFX = updateGFX
M.update = update

local function init()
	nd_tr1=nil
	nd_tr2=nil
	nd_tr3=nil
	electrics.values['custom_throttle'] = 0
	electrics.values['pitch_input'] = 0
	for k, node in pairs (v.data.nodes) do
		if v.data.nodes[node.cid].name == "pm1" then nd_tr1 = node.cid  end
		if v.data.nodes[node.cid].name == "pa1" then nd_tr2 = node.cid  end
		if v.data.nodes[node.cid].name == "pb1" then nd_tr3 = node.cid  end
		if v.data.nodes[node.cid].name == "pc1" then nd_tr4 = node.cid  end
	end
    if v.data.hydros == nil or (v.data.hydros ~= nil and tableSize(v.data.hydros) == 0) then
        M.updateGFX = nop
        M.update = nop
        M.hydros = {}
        return
    end

    M.hydros = shallowcopy(v.data.hydros)

    local hydroCount = 0
    for i, h in pairs (M.hydros) do
        hydroCount = hydroCount + 1
        h.inputCenter = h.inputCenter * h.inputFactor
        h.inputInLimit = h.inputInLimit * h.inputFactor
        h.inputOutLimit = h.inputOutLimit * h.inputFactor
        local inputFactorSign = sign(h.inputFactor)

        if h.inputFactor < 0 then
            h.inputInLimit, h.inputOutLimit = h.inputOutLimit, h.inputInLimit
        end

        local inputMiddle = (h.inputOutLimit + h.inputInLimit) * 0.5
        if h.inputCenter >= inputMiddle then
            h.center = 1 + (h.outLimit - 1) * (h.inputCenter - inputMiddle) / (h.inputOutLimit - inputMiddle)
        else
            h.center = 1 - (1 - h.inLimit) * (inputMiddle - h.inputCenter) / (inputMiddle - h.inputInLimit)
        end
        h.state = h.center
        h.multOut = (h.outLimit - h.center) / (h.inputOutLimit - h.inputCenter)
        h.cOut = h.center - h.inputCenter * h.multOut
        h.multIn = (h.center - h.inLimit) / (h.inputCenter - h.inputInLimit)
        h.cIn = h.center - h.inputCenter * h.multIn
        h.cmd = h.inputCenter
        h.invMultOut = 1 / (h.outLimit - h.center) * inputFactorSign
        h.invMultIn = 1 / (h.center - h.inLimit) * inputFactorSign
        h._inrate = h.inRate
        h._outrate = h.outRate
        h.hydroDirState = 0
        
        -- little workaround
        if h.inputSource == "steering" then
            h.inputSource = "steering_input"
        end

        if h.inputSource == "steering_input" then
            h.steering = true
        end
    end

    if hydroCount == 0 then
        M.updateGFX = nop
        M.update = nop 
    end

    FFBHydros = {}
    for i,h in pairs(M.hydros) do
        if h.steering then
            h.invFFBHydroRefL = 1 / obj:getBeamRefLength(h.beam.cid)
            table.insert(FFBHydros, h)
        end
    end

    if #FFBHydros ~= 0 then
        FFBHydrosExist = true
        FFBsmooth:set(0)
        curForceLimitSmoother:set(0)
        FFBIntegral = 0

        if M.enableFFB and FFBID < 0 then
            FFBID = obj:setupForceFeedback("xaxis")
        end

        if FFBID >= 0 then
            obj:sendForceFeedback(FFBID, 0)
            log('I', "hydros.init", "yay, force feedback is available :D")

            local config_json = obj:getForceFeedbackConfig(FFBID)
            local state, config = pcall(json.decode, config_json)
            if state == false then
                log('W', "hydros.init", "unable to decode JSON: "..tostring(v))
                log('W', "hydros.init", "JSON decoding error: "..tostring(config))
            else
                log('I', "hydros.init", "Force Feedback Configuration:")
                -- dump(config)
                local FFaxis = config["controls"]["xaxis"]
                if FFaxis ~= nil and FFaxis.ff_max_force ~= 0 then
                    FFmax = FFaxis.ff_max_force
                    M.wheelFFBForceLimit = math.min(M.wheelFFBForceLimit, FFmax)
                    FFstep = FFmax / FFaxis.ff_res
                    dump(FFaxis)
                    local state, ffbparams = pcall(json.decode, FFaxis["ffbParams"])
                    if state == true then
                        if ffbparams[  "forceCoef"] ~= nil then M.wheelFFBForceCoef   = ffbparams["forceCoef"]   end
                        if ffbparams[ "forceLimit"] ~= nil then M.wheelFFBForceLimit  = ffbparams["forceLimit"]  end
                        if ffbparams[  "smoothing"] ~= nil then M.wheelFFBSmoothing   = ffbparams["smoothing"]   end
                        if ffbparams["smoothingHF"] ~= nil then M.wheelFFBSmoothingHF = ffbparams["smoothingHF"] end
                    end
                end
            end
            log("I", "hydros.init", "FFB parameters are wheelFFBForceCoef: "..M.wheelFFBForceCoef..", wheelFFBForceLimit: "..M.wheelFFBForceLimit..", wheelFFBSmoothing: "..M.wheelFFBSmoothing..", wheelFFBSmoothingHF: "..M.wheelFFBSmoothingHF)

        end
    end
end

local function mapsReloaded()
    FFBID = -1
    init()
end

local function reset()
    electrics.values['custom_throttle']=0
    for k,h in pairs(M.hydros) do
        h.state = h.center
        h.cmd = h.inputCenter
        h._inrate = h.inRate
        h._outrate = h.outRate
    end

    FFBsmooth:set(0)
    curForceLimitSmoother:set(0)
    FFBIntegral = 0
    if FFBID >= 0 then
        obj:sendForceFeedback(FFBID, 0)
    end
end

local function sendHydroStateToGUI()
    obj:executeJS("HookManager.trigger('HydrosUpdate', "..encodeJson(M.hydros)..");")
end

-- public interface
M.init = init
M.reset = reset
M.sendHydroStateToGUI = sendHydroStateToGUI
M.mapsReloaded = mapsReloaded

return M